from argparse import ArgumentParser, BooleanOptionalAction, Namespace
from os.path import expanduser
from sys import stdout
from uuid import UUID

from .cli import CLI
from .version import VERSION


def parse_args() -> Namespace:
    parser = ArgumentParser(
        prog="Safari Bookmarks CLI",
        description="A utility to help manage Safari bookmarks.",
    )
    parser.set_defaults(command=None)
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s v{version}".format(version=VERSION),
    )
    parser.add_argument(
        "--file", "-f",
        type=expanduser,
        default="~/Library/Safari/Bookmarks.plist",
        help="The path to the Safari bookmarks.",
    )
    parser.add_argument(
        "--json",
        action=BooleanOptionalAction,
        default=False,
        help="Render output as JSON",
    )
    parser.add_argument(
        "--format", "-F",
        required=False,
        help="Customize the output format. Available placeholders: {title}, {url}, {id}, {type}, {prefix}, {suffix}."
    )
    subparsers = parser.add_subparsers(
        title="commands",
        required=True,
    )
    parser_list = subparsers.add_parser(
        "list",
        aliases=["ls", "show"],
        description="List bookmarks and folders.",
    )
    parser_list.add_argument(
        "target",
        nargs="?",
        help="The UUID or title of the bookmark or folder to show. Default shows all.",
    )
    parser_list.set_defaults(command="list")
    parser_add = subparsers.add_parser(
        "add",
        aliases=["a", "create"],
        description="Add a bookmark or folder.",
    )
    parser_add.add_argument(
        "title",
        help="The title of the bookmark or folder.",
    )
    parser_add.add_argument(
        "--uuid",
        type=UUID,
        required=False,
        help="The UUID to use. Default is to generate a new UUID.",
    )
    parser_add.add_argument(
        "--to",
        help="The target folder.",
    )
    group = parser_add.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--url",
        help="The URL for the bookmark."
    )
    group.add_argument(
        "--folder",
        dest="list",
        action=BooleanOptionalAction,
        default=False,
        help="Add a folder instead of a bookmark.",
    )
    parser_add.set_defaults(command="add")
    parser_remove = subparsers.add_parser(
        "remove",
        aliases=["rm", "delete", "del"],
        description="Remove a bookmark or folder.",
    )
    parser_remove.add_argument(
        "target",
        help="The UUID or title of the bookmark or folder to remove."
    )
    parser_remove.set_defaults(command="remove")
    parser_move = subparsers.add_parser(
        "move",
        aliases=["mv"],
        description="Move a bookmark or folder.",
    )
    parser_move.add_argument(
        "target",
        help="The UUID or title of the bookmark or folder to move."
    )
    parser_move.add_argument(
        "--to",
        help="The destination folder.",
    )
    parser_move.set_defaults(command="move")
    parser_move = subparsers.add_parser(
        "edit",
        aliases=["e", "update", "change"],
        description="Edit a bookmark or folder.",
    )
    parser_move.add_argument(
        "target",
        help="The UUID or title of the bookmark or folder to change.",
    )
    parser_move.add_argument(
        "--title",
        help="The new title to change.",
    )
    parser_move.add_argument(
        "--url",
        help="The new URL to change. Only used for bookmarks.",
    )
    parser_move.set_defaults(command="move")
    return parser.parse_args()


def main():
    args = parse_args()
    CLI(args.file, stdout).run(args.command, args)



import plistlib
from plistlib import PlistFormat
from typing import IO, Optional
from .models import WebBookmarkTypeList


def load(fp: IO, fmt: Optional[PlistFormat] = None) -> WebBookmarkTypeList:
    if hasattr(fp, "mode"):
        assert 'b' in fp.mode
    data = plistlib.load(fp, fmt=fmt)
    return WebBookmarkTypeList.model_validate(data)


def dump(bookmarks: WebBookmarkTypeList, fp: IO, fmt: PlistFormat) -> None:
    if hasattr(fp, "mode"):
        assert 'b' in fp.mode
    data = bookmarks.model_dump(by_alias=True)
    return plistlib.dump(data, fp, fmt=fmt, sort_keys=True, skipkeys=False)



from pydantic import BaseModel, Field
from typing import Annotated, Iterable, Literal, Union

ChildrenType = Annotated[
    Union["WebBookmarkTypeProxy", "WebBookmarkTypeLeaf", "WebBookmarkTypeList"],
    Field(discriminator="web_bookmark_type"),
]


class CloudKitSync(BaseModel, populate_by_name=True):
    cloud_kit_device_identifier: str = Field(alias="CloudKitDeviceIdentifier")
    cloud_kit_migration_state: int = Field(alias="CloudKitMigrationState")


class WebBookmarkType(BaseModel, extra="allow", populate_by_name=True):
    web_bookmark_uuid: str = Field(alias="WebBookmarkUUID")
    web_bookmark_type: str = Field(alias="WebBookmarkType")

    def __hash__(self) -> int:
        return hash(self.web_bookmark_uuid)


class WebBookmarkTypeProxy(WebBookmarkType):
    web_bookmark_type: Literal["WebBookmarkTypeProxy"] = Field(alias="WebBookmarkType", default="WebBookmarkTypeProxy")
    title: str = Field(alias="Title")


class WebBookmarkTypeLeaf(WebBookmarkType):
    web_bookmark_type: Literal["WebBookmarkTypeLeaf"] = Field(alias="WebBookmarkType", default="WebBookmarkTypeLeaf")
    url_string: str = Field(alias="URLString")
    uri_dictionary: dict[str, str] = Field(alias="URIDictionary", default_factory=dict)


class WebBookmarkTypeList(WebBookmarkType):
    web_bookmark_type: Literal["WebBookmarkTypeList"] = Field(alias="WebBookmarkType", default="WebBookmarkTypeList")
    title: str = Field(alias="Title")
    children: list[ChildrenType] = Field(alias="Children", default_factory=list)

    def __iter__(self) -> Iterable[WebBookmarkType]:
        return iter(self.children)

    def append(self, item: WebBookmarkType) -> None:
        self.children.append(item)

    def insert(self, index: int, item: WebBookmarkType) -> None:
        self.children.insert(index, item)

    def remove(self, item: WebBookmarkType) -> None:
        self.children.remove(item)



from contextlib import contextmanager
import plistlib
from typing import Generator, IO, Optional
import uuid
from .helpers import load, dump
from .models import WebBookmarkType, WebBookmarkTypeList, WebBookmarkTypeLeaf, WebBookmarkTypeProxy

DEFAULT_LIST_FORMAT = "{prefix: <{depth}}{title: <50}{type: <6}{id: <38}{url}"


class CLI():
    def __init__(self, path: str, out: IO) -> None:
        self.path = path
        self.output = out

    def run(self, command, args) -> None:
        if command is None:
            raise ValueError("No command specified")
        if func := getattr(self, command):
            func(args)
        else:
            raise ValueError(f"Invalid command: {command}")

    @contextmanager
    def with_bookmarks(self, update: bool = False) -> Generator[WebBookmarkTypeList, None, None]:
        with open(self.path, "rb") as file:
            bookmarks = load(file, plistlib.FMT_BINARY)
        yield bookmarks
        if update:
            with open(self.path, "wb") as file:
                dump(bookmarks, file, plistlib.FMT_BINARY)

    def lookup(self, title: str, root: WebBookmarkType) -> Optional[WebBookmarkType]:
        if title.lower() == str(root.web_bookmark_uuid).lower():
            return root
        elif isinstance(root, WebBookmarkTypeLeaf):
            if root.uri_dictionary.get("title") == title:
                return root
        elif root.title == title:
            return root
        elif isinstance(root, WebBookmarkTypeList):
            for child in root:
                if result := self.lookup(title, child):
                    return result
        return None

    def parent(self, target: WebBookmarkType, root: WebBookmarkType) -> Optional[WebBookmarkTypeList]:
        if target == root:
            return root
        elif isinstance(root, WebBookmarkTypeList):
            if target in root:
                return root
            for child in root:
                if result := self.parent(target, child):
                    return result
        return None

    def get_info(self, item: WebBookmarkType) -> tuple[str, str, str]:
        if isinstance(item, WebBookmarkTypeLeaf):
            return (
                "leaf",
                item.uri_dictionary.get("title", ""),
                item.url_string,
            )
        elif isinstance(item, WebBookmarkTypeProxy):
            return (
                "proxy",
                item.title,
                ""
            )
        elif isinstance(item, WebBookmarkTypeList):
            return (
                "list",
                item.title,
                "",
            )
        else:
            return ("unknown", "", "")

    def render_item(self, item: WebBookmarkType, format: str, depth: int = 0):
        id = item.web_bookmark_uuid
        type_, title, url = self.get_info(item)
        self.output.write(
            f"{format}\n".format(
                depth=depth,
                prefix="",
                suffix="",
                title=title.replace("\n", ""),
                type=type_,
                url=url,
                id=str(id),
            )
        )
        if isinstance(item, WebBookmarkTypeList):
            self.render_children(item, format=format, depth=depth+1)
            self.output.write("\n")

    def render_children(self, item: WebBookmarkType, format: str, depth: int = 0):
        for child in item.children:
            self.render_item(child, format, depth=depth)

    def render(self, root: WebBookmarkType, args, only_children=False):
        if args.json:
            self.output.write(root.model_dump_json(by_alias=True))
        else:
            format = args.format if args.format else DEFAULT_LIST_FORMAT
            if only_children:
                self.render_children(root, format=format)
            else:
                self.render_item(root, format=format)

    def list(self, args):
        with self.with_bookmarks("rb") as target:
            if args.target:
                target = self.lookup(args.target, target)
            if target is None:
                raise ValueError("Target not found")
            self.render(target, args, True)

    def add(self, args):
        uuid_ = str(args.uuid or uuid.uuid4()).upper()
        if args.list:
            if not args.title:
                raise ValueError("Title is required")
            web_bookmark = WebBookmarkTypeList(
                web_bookmark_uuid=uuid_,
                title=args.title,
            )
            if args.url:
                raise ValueError("URL is not supported by lists")
        else:
            web_bookmark = WebBookmarkTypeLeaf(
                web_bookmark_uuid=uuid_,
                url_string=args.url,
            )
            if args.title:
                web_bookmark.uri_dictionary["title"] = args.title
        with self.with_bookmarks(True) as target:
            if args.to:
                target = self.lookup(args.to, target)
                if not isinstance(target, WebBookmarkTypeList):
                    raise ValueError("Invalid destination")
            target.children.append(web_bookmark)
            self.render(web_bookmark, args)

    def remove(self, args):
        with self.with_bookmarks(True) as parent:
            target = self.lookup(args.target, parent)
            if target is None:
                raise ValueError("Target not found")
            parent = self.parent(target, parent)
            parent.remove(target)
            self.render(target, args)

    def move(self, args):
        with self.with_bookmarks(True) as bookmarks:
            target = self.lookup(args.target, bookmarks)
            if target is None:
                raise ValueError("Target not found")
            parent = self.parent(target, bookmarks)
            if args.to:
                dest = self.lookup(args.to, bookmarks)
            if not isinstance(dest, WebBookmarkTypeList):
                raise ValueError("Invalid destination")
            parent.remove(target)
            dest.append(target)
            self.render(dest, args)

    def edit(self, args):
        with self.with_bookmarks(True) as bookmarks:
            target = self.lookup(args.target, bookmarks)
            if target is None:
                raise ValueError("Target not found")
            if title := args.title:
                if isinstance(target, WebBookmarkTypeList):
                    target.title = title
                elif isinstance(target, WebBookmarkTypeLeaf):
                    target.uri_dictionary["title"] = title
                else:
                    raise ValueError("Cannot update target title")
            if url := args.url:
                if isinstance(target, WebBookmarkTypeLeaf):
                    target.url_string = url
                else:
                    raise ValueError("Cannot update target url")
            self.render(target, args)
